use async_channel::{Receiver, Sender};
use async_std::task;
use gio::prelude::*;
use glib::subclass::prelude::ObjectSubclass;
use glib::subclass::prelude::*;
use glib::subclass::Signal;
use glib::{
    clone, ObjectExt, ParamFlags, ParamSpec, ParamSpecBoolean, ParamSpecObject, ParamSpecString,
    ParamSpecUInt, StaticType, ToValue,
};
use once_cell::sync::{Lazy, OnceCell};
use transmission_client::{Authentication, Client, ClientError, Torrent};
use url::Url;

use std::cell::RefCell;
use std::collections::BTreeMap;
use std::convert::TryInto;
use std::time::Duration;

use crate::{TrAuthentication, TrSession, TrSessionStats, TrTorrent, TrTorrentModel};

mod imp {
    use super::*;

    #[derive(Debug)]
    pub struct TrClient {
        pub address: RefCell<String>,
        pub polling_rate: RefCell<u64>,
        pub torrents: TrTorrentModel,
        pub session: OnceCell<TrSession>,
        pub session_stats: TrSessionStats,

        pub client: RefCell<Option<Client>>,
        pub authentication: RefCell<Option<TrAuthentication>>,

        pub do_connect: RefCell<bool>,
        pub do_disconnect: RefCell<bool>,

        pub sender: Sender<bool>,
        pub receiver: Receiver<bool>,
    }

    #[glib::object_subclass]
    impl ObjectSubclass for TrClient {
        const NAME: &'static str = "TrClient";
        type ParentType = glib::Object;
        type Type = super::TrClient;

        fn new() -> Self {
            let (sender, receiver) = async_channel::bounded(1);

            Self {
                address: RefCell::default(),
                polling_rate: RefCell::default(),
                torrents: TrTorrentModel::default(),
                session: OnceCell::default(),
                session_stats: TrSessionStats::default(),
                client: RefCell::default(),
                authentication: RefCell::default(),
                do_connect: RefCell::default(),
                do_disconnect: RefCell::default(),
                sender,
                receiver,
            }
        }
    }

    impl ObjectImpl for TrClient {
        fn signals() -> &'static [Signal] {
            static SIGNALS: Lazy<Vec<Signal>> = Lazy::new(|| {
                vec![
                    Signal::builder("connection-failure", &[], glib::Type::UNIT.into()).build(),
                    Signal::builder(
                        "torrent-added",
                        &[TrTorrent::static_type().into()],
                        glib::Type::UNIT.into(),
                    )
                    .build(),
                    Signal::builder(
                        "torrent-downloaded",
                        &[TrTorrent::static_type().into()],
                        glib::Type::UNIT.into(),
                    )
                    .build(),
                ]
            });
            SIGNALS.as_ref()
        }

        fn properties() -> &'static [ParamSpec] {
            static PROPERTIES: Lazy<Vec<ParamSpec>> = Lazy::new(|| {
                vec![
                    ParamSpecString::new(
                        "address",
                        "Address",
                        "Address",
                        None,
                        ParamFlags::READABLE,
                    ),
                    ParamSpecUInt::new(
                        "polling-rate",
                        "Polling rate",
                        "Polling rate",
                        0,
                        std::u32::MAX,
                        0,
                        ParamFlags::READABLE,
                    ),
                    ParamSpecObject::new(
                        "torrents",
                        "Torrents",
                        "Torrents",
                        TrTorrentModel::static_type(),
                        ParamFlags::READABLE,
                    ),
                    ParamSpecObject::new(
                        "session",
                        "Session",
                        "Session",
                        TrSession::static_type(),
                        ParamFlags::READABLE,
                    ),
                    ParamSpecObject::new(
                        "session-stats",
                        "Session Stats",
                        "Session Stats",
                        TrSessionStats::static_type(),
                        ParamFlags::READABLE,
                    ),
                    ParamSpecBoolean::new(
                        "is-busy",
                        "Is busy",
                        "Is busy",
                        false,
                        glib::ParamFlags::READABLE,
                    ),
                    ParamSpecBoolean::new(
                        "is-connected",
                        "Is Connected",
                        "Is Connected",
                        false,
                        glib::ParamFlags::READABLE,
                    ),
                ]
            });
            PROPERTIES.as_ref()
        }

        fn property(&self, _obj: &Self::Type, _id: usize, pspec: &ParamSpec) -> glib::Value {
            match pspec.name() {
                "address" => self.address.borrow().to_value(),
                "polling-rate" => self.polling_rate.borrow().to_value(),
                "torrents" => self.torrents.to_value(),
                "session" => self.session.get().unwrap().to_value(),
                "session-stats" => self.session_stats.to_value(),
                "is-busy" => (*self.do_connect.borrow() || *self.do_disconnect.borrow()).to_value(),
                "is-connected" => self.client.borrow().is_some().to_value(),
                _ => unimplemented!(),
            }
        }
    }
}

glib::wrapper! {
    pub struct TrClient(ObjectSubclass<imp::TrClient>);
}

impl TrClient {
    pub fn new() -> Self {
        let client: Self = glib::Object::new::<Self>(&[]).unwrap();

        let session = TrSession::new(&client);
        let imp = imp::TrClient::from_instance(&client);
        imp.session.set(session).unwrap();

        client
    }

    pub async fn test_connectivity(
        address: String,
        auth: Option<TrAuthentication>,
    ) -> Result<(), ClientError> {
        let url = Url::parse(&address).unwrap();
        let rpc_client = Client::new(url);

        if let Some(auth) = auth {
            let rpc_auth = Authentication {
                username: auth.username(),
                password: auth.password(),
            };
            rpc_client.set_authentication(Some(rpc_auth));
        }

        let session = rpc_client.session().await?;
        debug!(
            "Connectivity test to {} succeeded: Transmission daemon version {}",
            address, session.version
        );

        Ok(())
    }

    pub async fn test_port(&self) -> Result<bool, ClientError> {
        if let Some(rpc_client) = self.rpc_client() {
            rpc_client.port_test().await
        } else {
            warn!("Unable to test port, no rpc connection");
            Ok(false)
        }
    }

    pub async fn connect(&self, address: String, polling_rate: u64) -> Result<(), ClientError> {
        let imp = imp::TrClient::from_instance(&self);

        if self.is_busy() {
            warn!("Client is currently busy, unable to connect to new address.");
            return Ok(());
        }

        // With the do_connect/disconnect variables we avoid race conditions
        // eg. doing another connect attempt, while the client is already busy
        *imp.do_connect.borrow_mut() = true;
        self.notify("is-busy");

        // Do the actual connecting work
        let result = self.connect_internal(address, polling_rate).await;

        // Work is done -> unblock it again.
        *imp.do_connect.borrow_mut() = false;
        self.notify("is-busy");

        result
    }

    async fn connect_internal(
        &self,
        address: String,
        polling_rate: u64,
    ) -> Result<(), ClientError> {
        let imp = imp::TrClient::from_instance(&self);

        if self.is_connected() {
            self.disconnect(false).await;
        }

        debug!("Connect to {} ...", address);

        // Check if the address has changed, and delete cached auth information if needed
        if address != self.address() {
            *imp.authentication.borrow_mut() = None;
        }

        // Create new client for the rpc communication
        let url = Url::parse(&address).unwrap();
        let client = Client::new(url);

        // Authenticate if we have information...
        if imp.authentication.borrow().is_some() {
            self.rpc_auth(Some(client.clone()));
        }

        // Set properties
        *imp.address.borrow_mut() = address.clone();
        *imp.polling_rate.borrow_mut() = polling_rate;
        self.notify("address");
        self.notify("polling-rate");

        // Check server version
        let session = client.session().await?;
        debug!(
            "Connected to transmission session version {}",
            session.version
        );

        // Make first poll
        *imp.client.borrow_mut() = Some(client.clone());
        self.poll_data(true).await?;
        self.notify("is-connected");

        // Start polling
        let duration = std::time::Duration::from_millis(polling_rate);
        glib::timeout_add_local(
            duration,
            clone!(@weak self as this => @default-return glib::Continue(false), move ||{
                let imp = imp::TrClient::from_instance(&this);
                let disconnect = imp.do_disconnect.borrow().clone();

                let fut = clone!(@weak this, @strong disconnect => async move {
                    let imp = imp::TrClient::from_instance(&this);
                    if disconnect {
                        debug!("Stop polling loop...");

                        // Send message back to the disconnect method to indicate that the
                        // polling loop has been stopped
                        let sender = imp.sender.clone();
                        sender.send(true).await.unwrap();

                        *imp.do_disconnect.borrow_mut() = false;
                        this.notify("is-busy");
                    }else{
                        this.refresh_data().await;
                    }
                });
                spawn!(fut);

                glib::Continue(!disconnect)
            }),
        );

        Ok(())
    }

    pub async fn disconnect(&self, close_session: bool) {
        let imp = imp::TrClient::from_instance(&self);
        if !self.is_connected() {
            warn!("Unable to disconnect, is not connected.");
            return;
        }

        if close_session {
            let client = imp.client.borrow().as_ref().unwrap().clone();
            client.session_close().await.unwrap();
        }

        let _client = imp.client.borrow_mut().take().unwrap();
        self.notify("is-connected");

        // Wait till polling loop has stopped
        *imp.do_disconnect.borrow_mut() = true;
        self.notify("is-busy");

        // Wait from message from polling loop to ensure we're not polling anymore.
        let r = imp.receiver.recv().await.unwrap();
        debug!("Stopped polling: {:?}", r);

        // Not connected anymore -> clear torrents model
        imp.torrents.clear();

        debug!("Disconnected from transmission session");
    }

    pub fn set_authentication(&self, auth: TrAuthentication) {
        let imp = imp::TrClient::from_instance(&self);
        *imp.authentication.borrow_mut() = Some(auth);
        self.rpc_auth(self.rpc_client());
    }

    fn rpc_auth(&self, rpc_client: Option<Client>) {
        let imp = imp::TrClient::from_instance(&self);

        if let Some(auth) = &*imp.authentication.borrow() {
            if let Some(rpc_client) = rpc_client {
                let rpc_auth = Authentication {
                    username: auth.username(),
                    password: auth.password(),
                };
                rpc_client.set_authentication(Some(rpc_auth));
            } else {
                warn!("Unable to authenticate, no rpc connection");
            }
        } else {
            warn!("Unable to authenticate, no information stored");
        }
    }

    /// `filename` can be a magnet url or a local file path
    pub async fn add_torrent_by_filename(&self, filename: String) -> Result<(), ClientError> {
        if let Some(client) = self.rpc_client() {
            client.torrent_add_filename(&filename).await?;
            self.refresh_data().await;
        } else {
            warn!("Unable to add new torrent, no rpc connection");
        }

        Ok(())
    }

    /// `metainfo` is the base64 encoded content of a .torrent file
    pub async fn add_torrent_by_metainfo(&self, metainfo: String) -> Result<(), ClientError> {
        if let Some(client) = self.rpc_client() {
            client.torrent_add_metainfo(&metainfo).await?;
            self.refresh_data().await;
        } else {
            warn!("Unable to add new torrent, no rpc connection");
        }

        Ok(())
    }

    pub async fn remove_torrents(
        &self,
        only_downloaded: bool,
        delete_local_data: bool,
    ) -> Result<(), ClientError> {
        let model = self.torrents();
        let mut remove_hashes = vec![];

        for i in 0..model.n_items() {
            let torrent = model.item(i).unwrap().downcast::<TrTorrent>().unwrap();

            if !only_downloaded || torrent.size() == torrent.downloaded() {
                remove_hashes.push(torrent.hash());
            }
        }

        if let Some(client) = self.rpc_client() {
            client
                .torrent_remove(Some(remove_hashes), delete_local_data)
                .await?;
            self.refresh_data().await;
        } else {
            warn!("Unable to remove torrents, no rpc connection");
        }

        Ok(())
    }

    pub async fn start_torrents(&self) -> Result<(), ClientError> {
        let model = self.torrents();
        let mut start_hashes = vec![];

        // We need to get the hashes of the torrents to be able to start all of them
        for i in 0..model.n_items() {
            let torrent = model.item(i).unwrap().downcast::<TrTorrent>().unwrap();
            start_hashes.push(torrent.hash());
        }

        if let Some(client) = self.rpc_client() {
            client.torrent_start(Some(start_hashes), false).await?;
            
            self.refresh_data().await;
        } else {
            warn!("Unable to start torrents, no rpc connection")
        }

        Ok(())
    }

    pub async fn stop_torrents(&self) -> Result<(), ClientError> {
        let model = self.torrents();
        let mut stop_hashes = vec![];

        for i in 0..model.n_items() {
            let torrent = model.item(i).unwrap().downcast::<TrTorrent>().unwrap();
            stop_hashes.push(torrent.hash());
        }

        if let Some(client) = self.rpc_client() {
            client.torrent_stop(Some(stop_hashes)).await?;

            // The daemon needs a short time to stop the torrents
            task::sleep(Duration::from_millis(500)).await;
            self.refresh_data().await;
        } else {
            warn!("Unable to stop torrents, no rpc connection");
        }

        Ok(())
    }

    /// Polls the latest information without waiting for the next polling timeout
    pub async fn refresh_data(&self) {
        if let Err(err) = self.poll_data(false).await {
            warn!("Couldn't poll transmission data: {}", err.to_string());
            self.clone().disconnect(false).await;
            self.emit_by_name::<()>("connection-failure", &[]);
        }
    }

    async fn poll_data(&self, is_initial_poll: bool) -> Result<(), ClientError> {
        let imp = imp::TrClient::from_instance(self);

        if let Some(client) = self.rpc_client() {
            let torrents = client.torrents(None).await?;

            // Transmission assigns each torrent a "global" queue position, which means
            // *every* torrent has a queue position, ignoring the torrent status. Therefore
            // that position doesn't match the actual position of the download / seed queue,
            // so we determine the position and set it as as property of TrTorrent.
            let download_queue = Self::sorted_queue_by_status(&torrents, 3); // 3 -> DownloadWait
            let seed_queue = Self::sorted_queue_by_status(&torrents, 5); // 5 -> SeedWait

            // We have to find out which torrent isn't included anymore,
            // so we can remove it from the model too.
            let mut hashes_delta = imp.torrents.get_hashes();

            // Count downloaded torrents so we can use that value for TrSessionStats
            let mut downloaded_torrents = 0;

            for rpc_torrent in torrents {
                let hash = rpc_torrent.hash_string.clone();
                let download_queue_pos = Self::queue_pos(&download_queue, &rpc_torrent);
                let seed_queue_pos = Self::queue_pos(&seed_queue, &rpc_torrent);

                // Increase downloaded torrents count if it's SeedWait(5) or Seed(6)
                if rpc_torrent.status == 5 || rpc_torrent.status == 6 {
                    downloaded_torrents += 1;
                }

                if let Some(torrent) = imp.torrents.torrent_by_hash(hash.clone()) {
                    // Update the download / seed queue position
                    torrent.refresh_queue_positions(download_queue_pos, seed_queue_pos);

                    // Check if torrent status or "global" queue position changed
                    if torrent.status() as u32 != rpc_torrent.status as u32
                        || torrent.queue_position() != rpc_torrent.queue_position
                    {
                        // Update torrent properties
                        torrent.refresh_values(rpc_torrent);

                        // ... yes -> emit torrent model `status-changed` signal
                        imp.torrents.emit_by_name::<()>("status-changed", &[]);
                    } else {
                        // Update torrent properties, without emiting the status-changed signal
                        torrent.refresh_values(rpc_torrent);
                    }

                    let index = hashes_delta.iter().position(|x| *x == hash).unwrap();
                    hashes_delta.remove(index);
                } else {
                    debug!("Add new torrent: {}", rpc_torrent.name);

                    let torrent = TrTorrent::from_rpc_torrent(
                        rpc_torrent,
                        download_queue_pos,
                        seed_queue_pos,
                        self.clone(),
                    );
                    imp.torrents.add_torrent(&torrent);

                    if !is_initial_poll {
                        self.emit_by_name::<()>("torrent-added", &[&torrent]);
                    }
                }
            }

            // Remove the deltas from the model
            for hash in hashes_delta {
                let torrent = imp.torrents.torrent_by_hash(hash).unwrap();
                imp.torrents.remove_torrent(&torrent);
            }

            let rpc_session = client.session().await?;
            self.session().refresh_values(rpc_session);

            let rpc_session_stats = client.session_stats().await?;
            self.session_stats()
                .refresh_values(rpc_session_stats, downloaded_torrents);
        } else {
            warn!("Unable to poll transmission data, no rpc connection.");
        }

        Ok(())
    }

    pub(crate) fn rpc_client(&self) -> Option<Client> {
        let imp = imp::TrClient::from_instance(self);
        imp.client.borrow().clone()
    }

    pub fn address(&self) -> String {
        self.property("address")
    }

    pub fn polling_rate(&self) -> u64 {
        self.property("polling-rate")
    }

    pub fn torrents(&self) -> TrTorrentModel {
        self.property("torrents")
    }

    pub fn session(&self) -> TrSession {
        self.property("session")
    }

    pub fn session_stats(&self) -> TrSessionStats {
        self.property("session-stats")
    }

    pub fn is_busy(&self) -> bool {
        self.property("is-busy")
    }

    pub fn is_connected(&self) -> bool {
        self.property("is-connected")
    }

    fn sorted_queue_by_status(rpc_torrents: &Vec<Torrent>, status_code: i32) -> Vec<Torrent> {
        let mut download_queue: BTreeMap<u64, &Torrent> = BTreeMap::new();

        // Get all torrents with given status code
        for torrent in rpc_torrents {
            if torrent.status == status_code {
                download_queue.insert(torrent.queue_position.try_into().unwrap(), torrent);
            }
        }

        // Insert the sorted torrents into a vec
        let mut result = Vec::new();
        for torrent in download_queue {
            result.push(torrent.1.clone());
        }

        result
    }

    fn queue_pos(queue: &Vec<Torrent>, rpc_torrent: &Torrent) -> i32 {
        queue
            .iter()
            .position(|t| t == rpc_torrent)
            .map(|pos| pos as i32)
            .unwrap_or(-1)
    }
}
