use gio::prelude::FileExt;
use glib::subclass::prelude::ObjectSubclass;
use glib::subclass::prelude::*;
use glib::{
    clone, ObjectExt, ParamSpec, ParamSpecBoolean, ParamSpecEnum, ParamSpecInt, ParamSpecObject,
    ParamSpecString, StaticType, ToValue,
};
use once_cell::sync::Lazy;
use once_cell::sync::OnceCell;
use transmission_client::{Session, SessionMutator};

use std::cell::{Cell, RefCell};

use crate::{TrClient, TrEncryption};

mod imp {
    use super::*;

    #[derive(Debug, Default)]
    pub struct TrSession {
        pub client: OnceCell<TrClient>,

        pub version: RefCell<String>,
        pub download_dir: RefCell<Option<gio::File>>,
        pub start_added_torrents: Cell<bool>,
        pub encryption: Cell<TrEncryption>,
        pub incomplete_dir_enabled: Cell<bool>,
        pub incomplete_dir: RefCell<Option<gio::File>>,
        pub download_queue_enabled: Cell<bool>,
        pub download_queue_size: Cell<i32>,
        pub seed_queue_enabled: Cell<bool>,
        pub seed_queue_size: Cell<i32>,
        pub port_forwarding_enabled: Cell<bool>,
        pub peer_port_random_on_start: Cell<bool>,
        pub peer_port: Cell<i32>,
        pub peer_limit_global: Cell<i32>,
        pub peer_limit_per_torrent: Cell<i32>,
    }

    #[glib::object_subclass]
    impl ObjectSubclass for TrSession {
        const NAME: &'static str = "TrSession";
        type ParentType = glib::Object;
        type Type = super::TrSession;
    }

    impl ObjectImpl for TrSession {
        fn properties() -> &'static [ParamSpec] {
            static PROPERTIES: Lazy<Vec<ParamSpec>> = Lazy::new(|| {
                vec![
                    ParamSpecString::new(
                        "version",
                        "Version",
                        "Version",
                        None,
                        glib::ParamFlags::READABLE,
                    ),
                    ParamSpecObject::new(
                        "download-dir",
                        "Download directory",
                        "Download directory",
                        gio::File::static_type(),
                        glib::ParamFlags::READWRITE,
                    ),
                    ParamSpecBoolean::new(
                        "start-added-torrents",
                        "Start added torrents",
                        "Start added torrents",
                        false,
                        glib::ParamFlags::READWRITE,
                    ),
                    ParamSpecEnum::new(
                        "encryption",
                        "Encryption",
                        "Encryption",
                        TrEncryption::static_type(),
                        TrEncryption::default() as i32,
                        glib::ParamFlags::READWRITE,
                    ),
                    ParamSpecBoolean::new(
                        "incomplete-dir-enabled",
                        "Incomplete directory enabled",
                        "Incomplete directory enabled",
                        false,
                        glib::ParamFlags::READWRITE,
                    ),
                    ParamSpecObject::new(
                        "incomplete-dir",
                        "Incomplete directory",
                        "Incomplete directory",
                        gio::File::static_type(),
                        glib::ParamFlags::READWRITE,
                    ),
                    ParamSpecBoolean::new(
                        "download-queue-enabled",
                        "Download queue enabled",
                        "Download queue enabled",
                        false,
                        glib::ParamFlags::READWRITE,
                    ),
                    ParamSpecInt::new(
                        "download-queue-size",
                        "Download queue size",
                        "Download queue size",
                        1,
                        std::i32::MAX,
                        1,
                        glib::ParamFlags::READWRITE,
                    ),
                    ParamSpecBoolean::new(
                        "seed-queue-enabled",
                        "Seed queue enabled",
                        "Seed queue enabled",
                        false,
                        glib::ParamFlags::READWRITE,
                    ),
                    ParamSpecInt::new(
                        "seed-queue-size",
                        "Seed queue size",
                        "Seed queue size",
                        1,
                        std::i32::MAX,
                        1,
                        glib::ParamFlags::READWRITE,
                    ),
                    ParamSpecBoolean::new(
                        "port-forwarding-enabled",
                        "Port forwarding enabled",
                        "Port forwarding enabled",
                        false,
                        glib::ParamFlags::READWRITE,
                    ),
                    ParamSpecBoolean::new(
                        "peer-port-random-on-start",
                        "Random peer port on startup",
                        "Random peer port on startup",
                        false,
                        glib::ParamFlags::READWRITE,
                    ),
                    ParamSpecInt::new(
                        "peer-port",
                        "Peer port",
                        "Peer port",
                        1,
                        std::i32::MAX,
                        1,
                        glib::ParamFlags::READWRITE,
                    ),
                    ParamSpecInt::new(
                        "peer-limit-global",
                        "Peer limit global",
                        "Peer limit global",
                        1,
                        std::i32::MAX,
                        1,
                        glib::ParamFlags::READWRITE,
                    ),
                    ParamSpecInt::new(
                        "peer-limit-per-torrent",
                        "Peer limit per torrent",
                        "Peer limit per torrent",
                        1,
                        std::i32::MAX,
                        1,
                        glib::ParamFlags::READWRITE,
                    ),
                ]
            });
            PROPERTIES.as_ref()
        }

        fn property(&self, _obj: &Self::Type, _id: usize, pspec: &ParamSpec) -> glib::Value {
            match pspec.name() {
                "version" => self.version.borrow().to_value(),
                "download-dir" => self.download_dir.borrow().to_value(),
                "start-added-torrents" => self.start_added_torrents.get().to_value(),
                "encryption" => self.encryption.get().to_value(),
                "incomplete-dir-enabled" => self.incomplete_dir_enabled.get().to_value(),
                "incomplete-dir" => self.incomplete_dir.borrow().to_value(),
                "download-queue-enabled" => self.download_queue_enabled.get().to_value(),
                "download-queue-size" => self.download_queue_size.get().to_value(),
                "seed-queue-enabled" => self.seed_queue_enabled.get().to_value(),
                "seed-queue-size" => self.seed_queue_size.get().to_value(),
                "port-forwarding-enabled" => self.port_forwarding_enabled.get().to_value(),
                "peer-port-random-on-start" => self.peer_port_random_on_start.get().to_value(),
                "peer-port" => self.peer_port.get().to_value(),
                "peer-limit-global" => self.peer_limit_global.get().to_value(),
                "peer-limit-per-torrent" => self.peer_limit_per_torrent.get().to_value(),
                _ => unimplemented!(),
            }
        }

        fn set_property(
            &self,
            obj: &Self::Type,
            _id: usize,
            value: &glib::Value,
            pspec: &ParamSpec,
        ) {
            match pspec.name() {
                "download-dir" => obj.set_download_dir(value.get().unwrap()),
                "start-added-torrents" => obj.set_start_added_torrents(value.get().unwrap()),
                "encryption" => obj.set_encryption(value.get().unwrap()),
                "incomplete-dir-enabled" => obj.set_incomplete_dir_enabled(value.get().unwrap()),
                "incomplete-dir" => obj.set_incomplete_dir(value.get().unwrap()),
                "download-queue-enabled" => obj.set_download_queue_enabled(value.get().unwrap()),
                "download-queue-size" => obj.set_download_queue_size(value.get().unwrap()),
                "seed-queue-enabled" => obj.set_seed_queue_enabled(value.get().unwrap()),
                "seed-queue-size" => obj.set_seed_queue_size(value.get().unwrap()),
                "port-forwarding-enabled" => obj.set_port_forwarding_enabled(value.get().unwrap()),
                "peer-port-random-on-start" => {
                    obj.set_peer_port_random_on_start(value.get().unwrap())
                }
                "peer-port" => obj.set_peer_port(value.get().unwrap()),
                "peer-limit-global" => obj.set_peer_limit_global(value.get().unwrap()),
                "peer-limit-per-torrent" => obj.set_peer_limit_per_torrent(value.get().unwrap()),
                _ => unimplemented!(),
            }
        }
    }
}

glib::wrapper! {
    pub struct TrSession(ObjectSubclass<imp::TrSession>);
}

impl TrSession {
    pub(crate) fn new(client: &TrClient) -> Self {
        let session: Self = glib::Object::new(&[]).unwrap();

        let imp = imp::TrSession::from_instance(&session);
        imp.client.set(client.clone()).unwrap();

        session
    }

    pub fn version(&self) -> String {
        self.property("version")
    }

    pub fn download_dir(&self) -> Option<gio::File> {
        self.property("download-dir")
    }

    pub fn set_download_dir(&self, value: gio::File) {
        let imp = imp::TrSession::from_instance(&self);
        *imp.download_dir.borrow_mut() = Some(value.clone());

        let mutator = SessionMutator {
            download_dir: Some(value.path().unwrap()),
            ..Default::default()
        };
        self.mutate_session(mutator.clone(), "download_dir");
    }

    pub fn start_added_torrents(&self) -> bool {
        self.property("start-added-torrents")
    }

    pub fn set_start_added_torrents(&self, value: bool) {
        let imp = imp::TrSession::from_instance(&self);
        imp.start_added_torrents.set(value);

        let mutator = SessionMutator {
            start_added_torrents: Some(value),
            ..Default::default()
        };
        self.mutate_session(mutator.clone(), "start-added-torrents");
    }

    pub fn encryption(&self) -> TrEncryption {
        self.property("encryption")
    }

    pub fn set_encryption(&self, value: TrEncryption) {
        let imp = imp::TrSession::from_instance(&self);
        imp.encryption.set(value);

        let mutator = SessionMutator {
            encryption: Some(value.into()),
            ..Default::default()
        };
        self.mutate_session(mutator.clone(), "encryption");
    }

    pub fn incomplete_dir_enabled(&self) -> bool {
        self.property("incomplete-dir-enabled")
    }

    pub fn set_incomplete_dir_enabled(&self, value: bool) {
        let imp = imp::TrSession::from_instance(&self);
        imp.incomplete_dir_enabled.set(value);

        let mutator = SessionMutator {
            incomplete_dir_enabled: Some(value),
            ..Default::default()
        };
        self.mutate_session(mutator.clone(), "incomplete-dir-enabled");
    }

    pub fn incomplete_dir(&self) -> Option<gio::File> {
        self.property("incomplete-dir")
    }

    pub fn set_incomplete_dir(&self, value: gio::File) {
        let imp = imp::TrSession::from_instance(&self);
        *imp.incomplete_dir.borrow_mut() = Some(value.clone());

        let mutator = SessionMutator {
            incomplete_dir: Some(value.path().unwrap()),
            ..Default::default()
        };
        self.mutate_session(mutator.clone(), "incomplete-dir");
    }

    pub fn download_queue_enabled(&self) -> bool {
        self.property("download-queue-enabled")
    }

    pub fn set_download_queue_enabled(&self, value: bool) {
        let imp = imp::TrSession::from_instance(&self);
        imp.download_queue_enabled.set(value);

        let mutator = SessionMutator {
            download_queue_enabled: Some(value),
            ..Default::default()
        };
        self.mutate_session(mutator.clone(), "download-queue-enabled");
    }

    pub fn download_queue_size(&self) -> i32 {
        self.property("download-queue-size")
    }

    pub fn set_download_queue_size(&self, value: i32) {
        let imp = imp::TrSession::from_instance(&self);
        imp.download_queue_size.set(value);

        let mutator = SessionMutator {
            download_queue_size: Some(value),
            ..Default::default()
        };
        self.mutate_session(mutator.clone(), "download-queue-size");
    }

    pub fn seed_queue_enabled(&self) -> bool {
        self.property("seed-queue-enabled")
    }

    pub fn set_seed_queue_enabled(&self, value: bool) {
        let imp = imp::TrSession::from_instance(&self);
        imp.seed_queue_enabled.set(value);

        let mutator = SessionMutator {
            seed_queue_enabled: Some(value),
            ..Default::default()
        };
        self.mutate_session(mutator.clone(), "seed-queue-enabled");
    }

    pub fn seed_queue_size(&self) -> i32 {
        self.property("seed-queue-size")
    }

    pub fn set_seed_queue_size(&self, value: i32) {
        let imp = imp::TrSession::from_instance(&self);
        imp.seed_queue_size.set(value);

        let mutator = SessionMutator {
            seed_queue_size: Some(value),
            ..Default::default()
        };
        self.mutate_session(mutator.clone(), "seed-queue-size");
    }

    pub fn port_forwarding_enabled(&self) -> bool {
        self.property("por-forwarding-enabled")
    }

    pub fn set_port_forwarding_enabled(&self, value: bool) {
        let imp = imp::TrSession::from_instance(&self);
        imp.port_forwarding_enabled.set(value);

        let mutator = SessionMutator {
            port_forwarding_enabled: Some(value),
            ..Default::default()
        };
        self.mutate_session(mutator.clone(), "port-forwarding-enabled");
    }

    pub fn peer_port_random_on_start(&self) -> bool {
        self.property("peer-port-random-on-start")
    }

    pub fn set_peer_port_random_on_start(&self, value: bool) {
        let imp = imp::TrSession::from_instance(&self);
        imp.peer_port_random_on_start.set(value);

        let mutator = SessionMutator {
            peer_port_random_on_start: Some(value),
            ..Default::default()
        };
        self.mutate_session(mutator.clone(), "peer-port-random-on-start");
    }

    pub fn peer_port(&self) -> i32 {
        self.property("peer-port")
    }

    pub fn set_peer_port(&self, value: i32) {
        let imp = imp::TrSession::from_instance(&self);
        imp.peer_port.set(value);

        let mutator = SessionMutator {
            peer_port: Some(value),
            ..Default::default()
        };
        self.mutate_session(mutator.clone(), "peer-port");
    }

    pub fn peer_limit_global(&self) -> i32 {
        self.property("peer-limit-global")
    }

    pub fn set_peer_limit_global(&self, value: i32) {
        let imp = imp::TrSession::from_instance(&self);
        imp.peer_limit_global.set(value);

        let mutator = SessionMutator {
            peer_limit_global: Some(value),
            ..Default::default()
        };
        self.mutate_session(mutator.clone(), "peer-limit-global");
    }

    pub fn peer_limit_per_torrent(&self) -> i32 {
        self.property("peer-limit-per-torrent")
    }

    pub fn set_peer_limit_per_torrent(&self, value: i32) {
        let imp = imp::TrSession::from_instance(&self);
        imp.peer_limit_per_torrent.set(value);

        let mutator = SessionMutator {
            peer_limit_per_torrent: Some(value),
            ..Default::default()
        };
        self.mutate_session(mutator.clone(), "peer-limit-per-torrent");
    }

    pub(crate) fn refresh_values(&self, rpc_session: Session) {
        let imp = imp::TrSession::from_instance(self);

        // version
        if *imp.version.borrow() != rpc_session.version {
            *imp.version.borrow_mut() = rpc_session.version;
            self.notify("version");
        }

        // download_dir
        let download_dir = gio::File::for_path(rpc_session.download_dir);
        if download_dir.path()
            != imp
                .download_dir
                .borrow()
                .as_ref()
                .map(|f| f.path())
                .flatten()
        {
            *imp.download_dir.borrow_mut() = Some(download_dir);
            self.notify("download-dir");
        }

        // start_added_torrents
        if imp.start_added_torrents.get() != rpc_session.start_added_torrents {
            imp.start_added_torrents
                .set(rpc_session.start_added_torrents);
            self.notify("start-added-torrents");
        }

        // encryption
        if imp.encryption.get() != rpc_session.encryption.clone().into() {
            imp.encryption.set(rpc_session.encryption.into());
            self.notify("encryption");
        }

        // incomplete_dir_enabled
        if imp.incomplete_dir_enabled.get() != rpc_session.incomplete_dir_enabled {
            imp.incomplete_dir_enabled
                .set(rpc_session.incomplete_dir_enabled.into());
            self.notify("incomplete-dir-enabled");
        }

        // incomplete_dir
        let incomplete_dir = gio::File::for_path(rpc_session.incomplete_dir);
        if incomplete_dir.path()
            != imp
                .incomplete_dir
                .borrow()
                .as_ref()
                .map(|f| f.path())
                .flatten()
        {
            *imp.incomplete_dir.borrow_mut() = Some(incomplete_dir);
            self.notify("incomplete-dir");
        }

        // download_queue_enabled
        if imp.download_queue_enabled.get() != rpc_session.download_queue_enabled {
            imp.download_queue_enabled
                .set(rpc_session.download_queue_enabled);
            self.notify("download-queue-enabled");
        }

        // download_queue_size
        if imp.download_queue_size.get() != rpc_session.download_queue_size.clone() {
            imp.download_queue_size.set(rpc_session.download_queue_size);
            self.notify("download-queue-size");
        }

        // seed_queue_enabled
        if imp.seed_queue_enabled.get() != rpc_session.seed_queue_enabled {
            imp.seed_queue_enabled.set(rpc_session.seed_queue_enabled);
            self.notify("seed-queue-enabled");
        }

        // seed_queue_size
        if imp.seed_queue_size.get() != rpc_session.seed_queue_size.clone() {
            imp.seed_queue_size.set(rpc_session.seed_queue_size);
            self.notify("seed-queue-size");
        }

        // port_forwarding_enabled
        if imp.port_forwarding_enabled.get() != rpc_session.port_forwarding_enabled {
            imp.port_forwarding_enabled
                .set(rpc_session.port_forwarding_enabled);
            self.notify("port-forwarding-enabled");
        }

        // peer_port_random_on_start
        if imp.peer_port_random_on_start.get() != rpc_session.peer_port_random_on_start {
            imp.peer_port_random_on_start
                .set(rpc_session.peer_port_random_on_start);
            self.notify("peer-port-random-on-start");
        }

        // peer_port
        if imp.peer_port.get() != rpc_session.peer_port.clone() {
            imp.peer_port.set(rpc_session.peer_port);
            self.notify("peer-port");
        }

        // peer_limit_global
        if imp.peer_limit_global.get() != rpc_session.peer_limit_global.clone() {
            imp.peer_limit_global.set(rpc_session.peer_limit_global);
            self.notify("peer-limit-global");
        }

        // peer_limit_per_torrent
        if imp.peer_limit_per_torrent.get() != rpc_session.peer_limit_per_torrent.clone() {
            imp.peer_limit_per_torrent
                .set(rpc_session.peer_limit_per_torrent);
            self.notify("peer-limit-per-torrent");
        }
    }

    fn mutate_session(&self, mutator: SessionMutator, prop_name: &str) {
        self.notify(prop_name);

        let fut = clone!(@weak self as this => async move {
            let imp = imp::TrSession::from_instance(&this);
            if let Some(rpc_client) = imp.client.get().unwrap().rpc_client() {
                rpc_client.session_set(mutator).await.unwrap();
            } else {
                warn!("Unable set mutate session, no rpc connection.");
            }
        });
        spawn!(fut);
    }
}
