use async_std::task;
use gio::prelude::*;
use gio::File;
use glib::subclass::prelude::ObjectSubclass;
use glib::subclass::prelude::*;
use glib::{
    ObjectExt, ParamSpec, ParamSpecBoolean, ParamSpecEnum, ParamSpecFloat, ParamSpecInt,
    ParamSpecInt64, ParamSpecString, StaticType, ToValue,
};
use once_cell::sync::Lazy;
use once_cell::sync::OnceCell;
use transmission_client::{ClientError, Torrent, TorrentMutator};

use std::cell::RefCell;
use std::convert::TryFrom;
use std::time::Duration;

use crate::client::TrClient;
use crate::torrent_status::TrTorrentStatus;

mod imp {
    use super::*;

    #[derive(Debug, Default)]
    pub struct TrTorrent {
        pub client: OnceCell<TrClient>,

        /* Static values */
        pub name: OnceCell<String>,
        pub hash: OnceCell<String>,
        pub primary_mime_type: OnceCell<String>,
        pub magnet_link: OnceCell<String>,

        /* Dynamic values */
        pub download_dir: RefCell<String>,
        pub size: RefCell<i64>,
        pub status: RefCell<TrTorrentStatus>,
        pub is_stalled: RefCell<bool>,
        pub eta: RefCell<i64>,
        pub error: RefCell<i32>,
        pub error_string: RefCell<String>,
        pub progress: RefCell<f32>,
        pub queue_position: RefCell<i32>,
        pub download_queue_position: RefCell<i32>,
        pub seed_queue_position: RefCell<i32>,

        pub seeders_active: RefCell<i32>,
        pub seeders: RefCell<i32>,
        pub leechers: RefCell<i32>,

        pub downloaded: RefCell<i64>,
        pub uploaded: RefCell<i64>,
        pub download_speed: RefCell<i32>,
        pub upload_speed: RefCell<i32>,
    }

    #[glib::object_subclass]
    impl ObjectSubclass for TrTorrent {
        const NAME: &'static str = "TrTorrent";
        type ParentType = glib::Object;
        type Type = super::TrTorrent;
    }

    impl ObjectImpl for TrTorrent {
        fn properties() -> &'static [ParamSpec] {
            static PROPERTIES: Lazy<Vec<ParamSpec>> = Lazy::new(|| {
                vec![
                    ParamSpecString::new("name", "Name", "Name", None, glib::ParamFlags::READABLE),
                    ParamSpecString::new("hash", "Hash", "Hash", None, glib::ParamFlags::READABLE),
                    ParamSpecString::new(
                        "primary-mime-type",
                        "Primary mime type",
                        "Primary mime type",
                        None,
                        glib::ParamFlags::READABLE,
                    ),
                    ParamSpecString::new(
                        "magnet-link",
                        "Magnet Link",
                        "Magnet Link",
                        None,
                        glib::ParamFlags::READABLE,
                    ),
                    ParamSpecString::new(
                        "download-dir",
                        "Download directory",
                        "Download directory",
                        None,
                        glib::ParamFlags::READABLE,
                    ),
                    ParamSpecInt64::new(
                        "size",
                        "Size",
                        "Size",
                        0,
                        i64::MAX,
                        0,
                        glib::ParamFlags::READABLE,
                    ),
                    ParamSpecEnum::new(
                        "status",
                        "Status",
                        "Status",
                        TrTorrentStatus::static_type(),
                        TrTorrentStatus::default() as i32,
                        glib::ParamFlags::READABLE,
                    ),
                    ParamSpecBoolean::new(
                        "is-stalled",
                        "Is stalled",
                        "Is stalled",
                        false,
                        glib::ParamFlags::READABLE,
                    ),
                    ParamSpecInt64::new(
                        "eta",
                        "ETA",
                        "ETA",
                        0,
                        i64::MAX,
                        0,
                        glib::ParamFlags::READABLE,
                    ),
                    ParamSpecInt::new(
                        "error",
                        "Error",
                        "Error",
                        0,
                        i32::MAX,
                        0,
                        glib::ParamFlags::READABLE,
                    ),
                    ParamSpecString::new(
                        "error-string",
                        "Error String",
                        "Error String",
                        None,
                        glib::ParamFlags::READABLE,
                    ),
                    ParamSpecFloat::new(
                        "progress",
                        "Progress",
                        "Progress",
                        0.0,
                        f32::MAX,
                        0.0,
                        glib::ParamFlags::READABLE,
                    ),
                    ParamSpecInt::new(
                        "queue-position",
                        "Queue position",
                        "Queue position",
                        0,
                        i32::MAX,
                        0,
                        glib::ParamFlags::READABLE,
                    ),
                    ParamSpecInt::new(
                        "download-queue-position",
                        "Download queue position",
                        "Download queue position",
                        0,
                        i32::MAX,
                        0,
                        glib::ParamFlags::READABLE,
                    ),
                    ParamSpecInt::new(
                        "seed-queue-position",
                        "Seed queue position",
                        "Seed queue position",
                        0,
                        i32::MAX,
                        0,
                        glib::ParamFlags::READABLE,
                    ),
                    ParamSpecInt::new(
                        "seeders-active",
                        "Seeders active",
                        "Seeders active",
                        0,
                        i32::MAX,
                        0,
                        glib::ParamFlags::READABLE,
                    ),
                    ParamSpecInt::new(
                        "seeders",
                        "Seeders",
                        "Seeders",
                        0,
                        i32::MAX,
                        0,
                        glib::ParamFlags::READABLE,
                    ),
                    ParamSpecInt::new(
                        "leechers",
                        "Leechers",
                        "Leechers",
                        0,
                        i32::MAX,
                        0,
                        glib::ParamFlags::READABLE,
                    ),
                    ParamSpecInt64::new(
                        "downloaded",
                        "Downloaded",
                        "Downloaded",
                        0,
                        i64::MAX,
                        0,
                        glib::ParamFlags::READABLE,
                    ),
                    ParamSpecInt64::new(
                        "uploaded",
                        "Uploaded",
                        "Uploaded",
                        0,
                        i64::MAX,
                        0,
                        glib::ParamFlags::READABLE,
                    ),
                    ParamSpecInt::new(
                        "download-speed",
                        "Download speed",
                        "Download speed",
                        0,
                        i32::MAX,
                        0,
                        glib::ParamFlags::READABLE,
                    ),
                    ParamSpecInt::new(
                        "upload-speed",
                        "Upload speed",
                        "Upload speed",
                        0,
                        i32::MAX,
                        0,
                        glib::ParamFlags::READABLE,
                    ),
                ]
            });
            PROPERTIES.as_ref()
        }

        fn property(&self, _obj: &Self::Type, _id: usize, pspec: &ParamSpec) -> glib::Value {
            match pspec.name() {
                "name" => self.name.get().unwrap().to_value(),
                "hash" => self.hash.get().unwrap().to_value(),
                "primary-mime-type" => self.primary_mime_type.get().unwrap().to_value(),
                "magnet-link" => self.magnet_link.get().unwrap().to_value(),
                "download-dir" => self.download_dir.borrow().to_value(),
                "size" => self.size.borrow().to_value(),
                "status" => self.status.borrow().to_value(),
                "is-stalled" => self.is_stalled.borrow().to_value(),
                "eta" => self.eta.borrow().to_value(),
                "error" => self.error.borrow().to_value(),
                "error-string" => self.error_string.borrow().to_value(),
                "progress" => self.progress.borrow().to_value(),
                "queue-position" => self.queue_position.borrow().to_value(),
                "download-queue-position" => self.download_queue_position.borrow().to_value(),
                "seed-queue-position" => self.seed_queue_position.borrow().to_value(),
                "seeders-active" => self.seeders_active.borrow().to_value(),
                "seeders" => self.seeders.borrow().to_value(),
                "leechers" => self.leechers.borrow().to_value(),
                "downloaded" => self.downloaded.borrow().to_value(),
                "uploaded" => self.uploaded.borrow().to_value(),
                "download-speed" => self.download_speed.borrow().to_value(),
                "upload-speed" => self.upload_speed.borrow().to_value(),
                _ => unimplemented!(),
            }
        }
    }
}

glib::wrapper! {
    pub struct TrTorrent(ObjectSubclass<imp::TrTorrent>);
}

impl TrTorrent {
    pub fn from_rpc_torrent(
        rpc_torrent: Torrent,
        download_queue_pos: i32,
        seed_queue_pos: i32,
        client: TrClient,
    ) -> Self {
        let torrent: Self = glib::Object::new(&[]).unwrap();
        let imp = imp::TrTorrent::from_instance(&torrent);

        imp.client.set(client).unwrap();

        // Set static values
        imp.hash.set(rpc_torrent.hash_string.clone()).unwrap();
        imp.name.set(rpc_torrent.name.clone()).unwrap();
        imp.primary_mime_type
            .set(rpc_torrent.primary_mime_type.clone())
            .unwrap();
        imp.magnet_link
            .set(rpc_torrent.magnet_link.clone())
            .unwrap();

        // Set dynamic values
        torrent.refresh_values(rpc_torrent);
        torrent.refresh_queue_positions(download_queue_pos, seed_queue_pos);

        torrent
    }

    pub async fn start(&self, bypass_queue: bool) -> Result<(), ClientError> {
        if let Some(client) = self.client().rpc_client() {
            client
                .torrent_start(Some(vec![self.hash()]), bypass_queue)
                .await?;
            self.do_poll().await;
        } else {
            warn!("Unable to start torrent, no rpc connection.");
        }

        Ok(())
    }

    pub async fn stop(&self) -> Result<(), ClientError> {
        if let Some(client) = self.client().rpc_client() {
            client.torrent_stop(Some(vec![self.hash()])).await?;

            // The daemon needs a short time to stop the torrent
            task::sleep(Duration::from_millis(500)).await;
            self.do_poll().await;
        } else {
            warn!("Unable to stop torrent, no rpc connection.");
        }

        Ok(())
    }

    pub async fn remove(&self, delete_local_data: bool) -> Result<(), ClientError> {
        if let Some(client) = self.client().rpc_client() {
            client
                .torrent_remove(Some(vec![self.hash()]), delete_local_data)
                .await?;
            self.do_poll().await;
        } else {
            warn!("Unable to stop torrent, no rpc connection.");
        }

        Ok(())
    }

    pub async fn reannounce(&self) -> Result<(), ClientError> {
        if let Some(client) = self.client().rpc_client() {
            client.torrent_reannounce(Some(vec![self.hash()])).await?;
            self.do_poll().await;
        } else {
            warn!("Unable to stop torrent, no rpc connection.");
        }

        Ok(())
    }

    pub async fn set_location(&self, location: File, move_data: bool) -> Result<(), ClientError> {
        let path = location.path().unwrap().to_str().unwrap().to_string();

        if let Some(client) = self.client().rpc_client() {
            client.torrent_set_location(Some(vec![self.hash()]), path, move_data).await?;
            self.do_poll().await;
        } else {
            warn!("Unable to set torrent location, no rpc connection.");
        }

        Ok(())
    }

    pub(crate) fn refresh_values(&self, rpc_torrent: Torrent) {
        let imp = imp::TrTorrent::from_instance(self);

        // download_dir
        if *imp.download_dir.borrow() != rpc_torrent.download_dir {
            *imp.download_dir.borrow_mut() = rpc_torrent.download_dir;
            self.notify("download-dir");
        }

        // size
        if *imp.size.borrow() != rpc_torrent.size_when_done {
            *imp.size.borrow_mut() = rpc_torrent.size_when_done;
            self.notify("size");
        }

        // status
        if *imp.status.borrow() as u32 != rpc_torrent.status as u32 {
            let status = TrTorrentStatus::try_from(rpc_torrent.status as u32).unwrap();

            // Check if status changed from "Downloading" to "Seeding"
            // and emit TrClient `torrent-downloaded` signal
            if self.status() == TrTorrentStatus::Download
                && (status == TrTorrentStatus::SeedWait || status == TrTorrentStatus::Seed)
            {
                imp.client
                    .get()
                    .unwrap()
                    .emit_by_name::<()>("torrent-downloaded", &[&self]);
            }

            *imp.status.borrow_mut() = status;
            self.notify("status");
        }

        // is_stalled
        if *imp.is_stalled.borrow() as u32 != rpc_torrent.is_stalled as u32 {
            *imp.is_stalled.borrow_mut() = rpc_torrent.is_stalled;
            self.notify("is-stalled");
        }

        // eta
        if *imp.eta.borrow() != rpc_torrent.eta {
            *imp.eta.borrow_mut() = rpc_torrent.eta;
            self.notify("eta");
        }

        // error
        if *imp.error.borrow() != rpc_torrent.error {
            *imp.error.borrow_mut() = rpc_torrent.error;
            self.notify("error");
        }

        // error-string
        if *imp.error_string.borrow() != rpc_torrent.error_string {
            *imp.error_string.borrow_mut() = rpc_torrent.error_string.clone();
            self.notify("error-string");
        }

        // progress
        if *imp.progress.borrow() != rpc_torrent.percent_done {
            *imp.progress.borrow_mut() = rpc_torrent.percent_done;
            self.notify("progress");
        }

        // queue_position
        if *imp.queue_position.borrow() != rpc_torrent.queue_position {
            *imp.queue_position.borrow_mut() = rpc_torrent.queue_position;
            self.notify("queue_position");
        }

        // seeders_active
        if *imp.seeders_active.borrow() != rpc_torrent.peers_sending_to_us {
            *imp.seeders_active.borrow_mut() = rpc_torrent.peers_sending_to_us;
            self.notify("seeders_active");
        }

        // seeders
        if *imp.seeders.borrow() != rpc_torrent.peers_connected {
            *imp.seeders.borrow_mut() = rpc_torrent.peers_connected;
            self.notify("seeders");
        }

        // leechers
        if *imp.leechers.borrow() != rpc_torrent.peers_getting_from_us {
            *imp.leechers.borrow_mut() = rpc_torrent.peers_getting_from_us;
            self.notify("leechers");
        }

        // downloaded
        if *imp.downloaded.borrow() != rpc_torrent.have_valid {
            *imp.downloaded.borrow_mut() = rpc_torrent.have_valid;
            self.notify("downloaded");
        }

        // uploaded
        if *imp.uploaded.borrow() != rpc_torrent.uploaded_ever {
            *imp.uploaded.borrow_mut() = rpc_torrent.uploaded_ever;
            self.notify("uploaded");
        }

        // downloaded
        if *imp.download_speed.borrow() != rpc_torrent.rate_download {
            *imp.download_speed.borrow_mut() = rpc_torrent.rate_download;
            self.notify("download-speed");
        }

        // uploaded
        if *imp.upload_speed.borrow() != rpc_torrent.rate_upload {
            *imp.upload_speed.borrow_mut() = rpc_torrent.rate_upload;
            self.notify("upload-speed");
        }
    }

    pub(crate) fn refresh_queue_positions(&self, download_queue_pos: i32, seed_queue_pos: i32) {
        let imp = imp::TrTorrent::from_instance(self);

        // download_queue_position
        if *imp.download_queue_position.borrow() != download_queue_pos {
            *imp.download_queue_position.borrow_mut() = download_queue_pos;
            self.notify("download-queue-position");
        }

        // seed_queue_position
        if *imp.seed_queue_position.borrow() != seed_queue_pos {
            *imp.seed_queue_position.borrow_mut() = seed_queue_pos;
            self.notify("seed-queue-position");
        }
    }

    pub fn name(&self) -> String {
        self.property("name")
    }

    pub fn hash(&self) -> String {
        self.property("hash")
    }

    pub fn primary_mime_type(&self) -> String {
        self.property("primary-mime-type")
    }

    pub fn magnet_link(&self) -> String {
        self.property("magnet-link")
    }

    pub fn download_dir(&self) -> String {
        self.property("download_dir")
    }

    pub fn size(&self) -> i64 {
        self.property("size")
    }

    pub fn status(&self) -> TrTorrentStatus {
        self.property("status")
    }

    pub fn is_stalled(&self) -> bool {
        self.property("is_stalled")
    }

    pub fn eta(&self) -> i64 {
        self.property("eta")
    }

    pub fn error(&self) -> i32 {
        self.property("error")
    }

    pub fn error_string(&self) -> String {
        self.property("error-string")
    }

    pub fn progress(&self) -> f32 {
        self.property("progress")
    }

    pub fn queue_position(&self) -> i32 {
        self.property("queue_position")
    }

    pub async fn set_queue_position(&self, pos: i32) -> Result<(), ClientError> {
        if let Some(client) = self.client().rpc_client() {
            let mutator = TorrentMutator {
                queue_position: Some(pos),
                ..Default::default()
            };
            client.torrent_set(Some(vec![self.hash()]), mutator).await?;
        } else {
            warn!("Unable set queue position, no rpc connection.");
        }

        Ok(())
    }

    pub fn download_queue_position(&self) -> i32 {
        self.property("download_queue_position")
    }

    pub async fn set_download_queue_position(&self, pos: i32) -> Result<(), ClientError> {
        let torrents = self.client().torrents();

        for i in 0..torrents.n_items() {
            let torrent: TrTorrent = torrents.item(i).unwrap().downcast().unwrap();
            if torrent.download_queue_position() == pos {
                self.set_queue_position(torrent.queue_position()).await?;
                break;
            }
        }
        self.do_poll().await;

        Ok(())
    }

    pub fn seed_queue_position(&self) -> i32 {
        self.property("seed_queue_position")
    }

    pub async fn set_seed_queue_position(&self, pos: i32) -> Result<(), ClientError> {
        let torrents = self.client().torrents();

        for i in 0..torrents.n_items() {
            let torrent: TrTorrent = torrents.item(i).unwrap().downcast().unwrap();
            if torrent.seed_queue_position() == pos {
                self.set_queue_position(torrent.queue_position()).await?;
                break;
            }
        }
        self.do_poll().await;

        Ok(())
    }

    pub fn seeders_active(&self) -> i32 {
        self.property("seeders_active")
    }

    pub fn seeders(&self) -> i32 {
        self.property("seeders")
    }

    pub fn leechers(&self) -> i32 {
        self.property("leechers")
    }

    pub fn downloaded(&self) -> i64 {
        self.property("downloaded")
    }

    pub fn uploaded(&self) -> i64 {
        self.property("uploaded")
    }

    pub fn download_speed(&self) -> i32 {
        self.property("download-speed")
    }

    pub fn upload_speed(&self) -> i32 {
        self.property("upload-speed")
    }

    fn client(&self) -> TrClient {
        let imp = imp::TrTorrent::from_instance(self);
        imp.client.get().unwrap().clone()
    }

    async fn do_poll(&self) {
        self.client().refresh_data().await;
    }
}

